library vasat;

import 'dart:convert';
import 'package:flutter/cupertino.dart';
import 'package:flutter_vasat/vasat_error.dart';
import 'package:http/http.dart' as http;
import 'package:shared_preferences/shared_preferences.dart';
import 'dart:async';
import 'models/vasat_model.dart';
import 'vasat_config.dart';
import 'models/session.dart';

/// Vasat flutter SDK
class Vasat {

  String name;
  VasatConfig config;
  Session _session;

  Session get currentSession => _session;

  static const String defaultName = "vasat";
  // _cache is library-private, thanks to the _ in front of its name.
  static final Map<String, Vasat> _cache = <String, Vasat>{};


  factory Vasat({String name = defaultName, VasatConfig config}) {
    // Return cached vasat for given name. if not present calls Vasat._init to generate one.
    return _cache.putIfAbsent(
        name, () => Vasat._init(name, config));
  }
  /// Additional method that does the same as the factory constructor. Just here for making it more natural to get instances of Vasat
  static Vasat instance([String name = defaultName]) {
    // Return cached vasat for given name. if not present calls Vasat._init to generate one.
    return _cache[name];
  }

  Vasat._init(this.name, this.config) {
    // initialization logic here
  }

  Future<bool> hasActiveSession() {
    return getActiveSession().then((session) {
      if (session != null && session.expiry == 0) {
        _session = session;
        return true;
      } else {
        _session = null;
        return false;
      }
    });
  }

  Future<Session> getActiveSession() {
    return getObjects<Session>(Session()).then((values) {
      if (values != null && values.length > 0) {
        return values[0];
      } else {
        return null;
      }
    });
  }

  Future<Session> loginWithClientCredentials() async {
    var body = jsonEncode(<String, String>{
      "grant_type": "client_credentials",
      "client_id": this.config?.clientId,
      "client_secret": this.config?.clientSecret,
    });
    return apiPost("/access/token", body).then((res) => processAccessToken(res));
  }

  Future<Session> loginWithUserCredentials(String username, String password) async {
    var body = jsonEncode(<String, String>{
      "grant_type": "password",
      "client_id": this.config?.clientId,
      "username": username,
      "password": password,
    });
    return apiPost("/access/token", body).then((res) => processAccessToken(res));
  }

  Future<dynamic> apiPost(String url, dynamic params, {bool returnPayloadOnly}) async {
    if (this.config != null && this.config.host != null) {
      return http.post(this.config.host + url, body: params, headers: getApiHeaders(isJson: true))
          .then((res) => processResponse(res, returnPayloadOnly: returnPayloadOnly));
    } else  {
      return Future.error("No vasat configuration or host found");
    }

  }

  Future<dynamic> apiMultipartPost(String url, dynamic params, List<String> filePaths) async {
    if (this.config != null && this.config.host != null) {
      String payload;
      if (params is String)
        payload = params;
      else if (params is Map)
        payload = jsonEncode(params);
      else
        payload = jsonEncode(params);

      var request = http.MultipartRequest("POST", Uri.parse(this.config.host + url))
          ..fields["payload"] = payload;

      for (var path in filePaths) {
        request.files.add(await getMultipartFile(path));
      }

      var res = await request.send();
      debug("Multipart response status: ${res.statusCode} stream: ${res.stream.toString()}");
    } else  {
      return Future.error("No vasat configuration or host found");
    }

  }

  Future<http.MultipartFile> getMultipartFile(String path) async {
    // infer content type
    String contentType;
    String fileName = path.substring(path.lastIndexOf("/")+1).toLowerCase();
    String ext = path.substring(path.lastIndexOf(".")+1).toLowerCase();
    switch(ext) {
      case "png":
        contentType = "application/octet-stream";
        break;
      case "jpg":
      case "jpeg":
        contentType = "application/octet-stream";
        break;
      default:
        contentType = "application/octet-stream";
        break;
    }
    //return http.MultipartFile.fromPath(fileName, path, contentType: MediaType.parse(contentType) );
    return http.MultipartFile.fromPath(fileName, path);
  }

  Future<dynamic> apiGet(String url, {Map<String, String> params, bool returnPayloadOnly}) async {
    if (this.config != null && this.config.host != null) {
      var uri = Uri.parse(this.config.host + url);
      if (params != null) uri = uri.replace(queryParameters: params);

      debug("apiGet URL: ${uri.toString()} With params: ${uri.queryParameters.toString()}");

      return http.get(uri, headers: getApiHeaders())
          .then((res) => processResponse(res, returnPayloadOnly: returnPayloadOnly));
    } else  {
      return Future.error("No vasat configuration or host found");
    }

  }

  Future<Session> processAccessToken(Map<String, dynamic> response) async {
    debug("processSession: Processing access token");
    _session = Session().fromJson(response);
    debug("processSession: Access token parsed");
    return saveObjects([_session]).then((value) => _session);
  }

  Map<String, String> getApiHeaders({bool isJson = false}) {
    var headers = <String,String>{};
    // If POST or PUT methods this should be set to true
    if (isJson) headers["Content-Type"] = "application/json";
    // Check if we have token
    if (_session?.token != null) headers["Authorization"] = "Bearer " + _session.token;

    return headers;
  }

  Future<dynamic> processResponse(http.Response response, {bool returnPayloadOnly = true}) async {

    // Always expect JSON payload. If not JSON throw an error.
    try {
      Map<String, dynamic> resBody = jsonDecode(response.body);
      debug("processResponse http status ${response.statusCode} response body: ${resBody.toString()}");
      //debugPrint("ENTER processResponse resBody:" + resBody.toString());

      if (response.statusCode < 400 && resBody.containsKey("access_token") ) {
        return Future<dynamic>.value(resBody);
      } else if (response.statusCode < 400 && resBody.containsKey("payload") && resBody.containsKey("status") && resBody["status"] < 400 && returnPayloadOnly) {
        return Future<dynamic>.value(resBody["payload"]);
      } else if (response.statusCode < 400 && resBody.containsKey("status") && resBody["status"] < 400 ) {
        return Future<dynamic>.value(resBody);
      } else  {
        return Future.error( VasatError.fromMap(resBody) );
      }
    } catch (error) {
      return Future.error(VasatError(800, "Api response error", error));
    }
  }

  Future<List<T>> syncFromRemote<T extends VasatModel>(T clazz, {Map<String, String> params}) async {

    var url = "/api/" + T.toString();
    debug("syncFromRemote URL: $url");
    var res = apiGet(url, params: params).then((value) {
      var results = value as List<dynamic>;
      debug("ENETR sync results: " + results.toString());
      return results.map((ele) => clazz.fromJson(ele) as T).toList();
    });

    return res;
  }

  Future<void> saveObjects<T extends VasatModel>(List<T> objects) async {
    debug("saveObjects");
    SharedPreferences prefs = await SharedPreferences.getInstance();
    var objs = objects.map((ele) {
      var obj = ele as VasatModel<T>;
      return jsonEncode( obj.toJson() );
    } ).toList();
    prefs.setStringList(T.toString(), objs);
  }

  Future<List<T>> getObjects<T extends VasatModel<T>>(T classType) async {

    return SharedPreferences.getInstance().then((prefs) {
      var list = prefs.getStringList(T.toString());
      List<T> objList;
      if (list != null) {
        objList = list.map((ele) {
          return classType.fromJson(jsonDecode(ele));
        } ).toList();
      }
      return objList;

    });
  }

  void debug(String message, { int wrapWidth }) {
    if (config.debug) debugPrint("Vasat - " + message, wrapWidth:wrapWidth);
  }
}
